<?php

namespace Myriad\EventDrivenService\Models;

use Carbon\Carbon;
use Illuminate\Database\Eloquent\SoftDeletes;
use Illuminate\Http\Resources\Json\JsonResource;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Gate;
use Illuminate\Support\Facades\Schema;
use Illuminate\Support\Facades\Validator;
use Illuminate\Support\Str;
use Illuminate\Validation\UnauthorizedException;
use Illuminate\Validation\ValidationException;
use Myriad\EventDrivenService\Interfaces\IUserFilterable;
use Myriad\EventDrivenService\Traits\Logger;
use Ramsey\Uuid\Uuid;

abstract class EventModel extends UUIDModel {

  use SoftDeletes, Logger;

  protected $commands = ['create', 'update', 'delete'];
  protected $resource = JsonResource::class;
  protected $allowRecreate = false;
  protected $nonColumnFields = [];

  protected $perPage = 1000;

  protected $ignoreOnClean = ['uuid'];

  public $modelKeyField = 'uuid';

  public $trackAccessChanges = false;

  public function getCommands() {
    return $this->commands;
  }

  public function canRecreate() {
    return $this->allowRecreate;
  }

  public function format() {
    return new $this->resource($this);
  }

  public function getUpdated($lastUpdate, $perPage) {
    $firstSync = !$lastUpdate || $lastUpdate == 0;
    if (!isset($perPage)) {
      $perPage = $this->getPerPage();
    }

    $query = $this->query()->select([$this->getTable() . '.*']);
    if (!$firstSync) {
      $lastUpdate = Carbon::createFromTimestamp($lastUpdate)->toDateTimeString();
      $query->updatedSince($lastUpdate)->withTrashed();
    }

    if ($this instanceof IUserFilterable) {
      $query->filterForUser();

      if ($this->trackAccessChanges && !$firstSync) {
        $columns = Schema::getColumnListing($this->getTable());

        $deletedQuery = $this->query()->select(array_merge(array_diff($columns, ['deleted_at']), [\DB::raw('NOW() as deleted_at')]))->withTrashed();

        if ($lastUpdate) {
          $deletedQuery->updatedSince($lastUpdate);
        }

        $filterQuery = $this->query()->select($this->getTable() . '.uuid')->filterForUser();

        $deletedQuery->whereRaw($this->getTable() . '.uuid not in (' .
          Str::replaceArray('?', $filterQuery->getBindings(), str_replace('?', "'?'", $filterQuery->toSql()))
          . ')');

        $deletedQuery->rightJoin('event_had_access', 'event_had_access.model_uuid', '=', $this->getTable() . '.uuid')
          ->where('event_had_access.user_uuid', Auth::user()->uuid);

        $query = $query->union($deletedQuery);

        $paginateQuery = $this->eagerLoad($this->query())->withTrashed()->fromSub($query, 'merged_objects')->select($columns);

        return $paginateQuery->orderBy('merged_objects.updated_at')->orderBy('merged_objects.uuid')->paginate($perPage);
      }
    }

    return $this->eagerLoad($query)->orderBy($this->getTable() . '.updated_at')->orderBy($this->getTable() . '.uuid')->paginate($perPage);
  }

  public function scopeWithAll($query) {
    return $query->withTrashed();
  }

  public function scopeUpdatedSince($query, $lastUpdate) {
    return $query->where($this->getTable() . '.updated_at', '>=', $lastUpdate);
  }

  public function needsFullSync($lastUpdate) {
    return false;
  }

  public function eagerLoad($query) {
    return $query;
  }

  public function getUpdatesResponse() {
    return [];
  }

  public function createModel($event) {
    if (!Gate::allows('create', [$this, $event['value']])) {
      throw new UnauthorizedException(__('event-driven-service::events.permission_create'));
    }
    $this->cleanEvent($event);

    $validator = Validator::make($event['value'], $this->getRules($event['value']));
    if ($validator->fails()) {
      throw new ValidationException($validator);
    }

    $this->fill($event['value']);
    $this->save();
  }

  public function updateModel($event) {
    if (!Gate::allows('update', [$this, $event])) {
      throw new UnauthorizedException(__('event-driven-service::events.permission_update'));
    }
    $values = array_merge($this->toArray(), ['value' => $event['value']]);
    $validator = Validator::make($values, ['value' => $this->getRules($values)[$event['name']] ?? []]);
    if ($validator->fails()) {
      throw new ValidationException($validator);
    }

    if (!in_array($event['name'], $this->nonColumnFields)) {
      $this->{$event['name']} = $event['value'];
      $this->{$event['name'] . '_version'} = $event['version_new'];
    }

    $this->save();
  }

  public function deleteModel($event) {
    if (!Gate::allows('delete', $this)) {
      throw new UnauthorizedException(__('event-driven-service::events.permission_delete'));
    }
    $this->delete();
  }

  public function isCurrent($name, $version) {
    $currentVersion = $this[$name . '_version'];
    return $currentVersion == null || $currentVersion == $version;
  }

  public function manuallyDelete() {
    return false;
  }

  protected function getRules($values) {
    return [];
  }

  public function getNonColumnFields() {
    return $this->nonColumnFields;
  }

  public function generateUuid() {
    return Uuid::uuid4();
  }

  protected function cleanEvent(&$event) {
    foreach ($event['value'] as $key => $value) {
      if (in_array($key, $this->ignoreOnClean)) {
        continue;
      }

      $column = substr($key, -8) != '_version' ? $key : substr($key, 0, strlen($key) - 8);
      $version = substr($key, -8) == '_version' ? $key : $key . '_version';

      if ((!array_key_exists($column, $event['value']) || !array_key_exists($version, $event['value'])) ||
        ($event['value'][$column] == null && $event['value'][$version] == null) ||
        !Uuid::isValid($event['value'][$version])) {
        unset($event['value'][$column]);
        unset($event['value'][$version]);
      }
    }
    return $event;
  }

}
