<?php

namespace Myriad\EventDrivenService\Services;

use Carbon\Carbon;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\ValidationException;
use Myriad\EventDrivenService\Http\Utilities\EventBuilder;
use Myriad\EventDrivenService\Http\Utilities\EventHelper;
use Myriad\EventDrivenService\Http\Utilities\Pagination;
use Myriad\EventDrivenService\Interfaces\IEventRepository;
use Myriad\EventDrivenService\Interfaces\IFailedEventRepository;
use Myriad\EventDrivenService\Models\EventHadAccess;
use Myriad\EventDrivenService\Models\EventModel;
use Myriad\EventDrivenService\Rules\UuidRule;
use Myriad\EventDrivenService\Traits\EventExceptionHandler;
use Ramsey\Uuid\Uuid;
use StableSort;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class EventService {

  use EventExceptionHandler;

  private $eventRepository;
  private $failedEventRepository;

  public function __construct(IEventRepository $eventRepository, IFailedEventRepository $failedEventRepository) {
    $this->eventRepository = $eventRepository;
    $this->failedEventRepository = $failedEventRepository;
  }

  public function getNew($uri, $lastUpdate, $pageSize) {
    $model = EventHelper::getModelByUri($uri);
    $modelType = EventHelper::getCommandForModel($model);

    $needSync = $model->needsFullSync($lastUpdate);
    $recordsFound = $model->getUpdated($lastUpdate, $pageSize);

    $currentData = [];
    foreach ($recordsFound as $record) {
      $event = EventBuilder::make($modelType, $record);
      if (property_exists($model, 'trackAccessChanges') && $model->trackAccessChanges)
        if ($event['command'] != $modelType . ':delete') {
          EventHadAccess::updateOrCreate(
            ['model_uuid' => $record->uuid, 'user_uuid' => Auth::user()->uuid, 'model' => $modelType],
            ['last_requested_at' => Carbon::now(), 'lost_at' => null]
          );
        } else if ($event['command'] == $modelType . ':delete') {
          EventHadAccess::updateOrCreate(
            ['model_uuid' => $record->uuid, 'user_uuid' => Auth::user()->uuid, 'model' => $modelType],
            ['lost_at' => Carbon::now()]
          );
        }
      $currentData[] = $event;
    }

    $meta = Pagination::build($recordsFound);
    $meta['full_sync_required'] = $needSync;

    return [
      'data' => collect($currentData),
      'meta' => $meta,
    ];
  }

  public function handleBulkEvents(Request $request) {
    try {
      DB::beginTransaction();
      list($errors) = $this->handlePost($request, false, true);
      if (count($errors) > 0) {
        DB::rollBack();
        $this->failedEventRepository->createAll($errors);
        return response()->json([
          'errors' => $errors->groupBy([function ($value, $key) {
            return explode(':', $value['request']['command'])[0];
          }, function ($value, $key) {
            return explode(':', $value['request']['command'])[1];
          }, function ($value, $key) {
            return explode(':', $value['request']['command'])[1] !== 'update' ? 'all' : @$value['request']['name'];
          }])->map(function ($commandLevel) {
            return $commandLevel->map(function ($fieldLevel, $command) {
              if ($command !== 'update') {
                return $fieldLevel->first()->first()['error'];
              }
              return $fieldLevel->map(function ($updateLevel) {
                return $updateLevel->first()['error'];
              });
            });
          })->toArray()
        ], 400);
      }
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
    return response()->json([], 201);
  }

  public function handlePost(Request $request, $processUpdateResponse, $isBulk = false) {
    list($valid, $invalid) = collect($request->data)
      ->partition(function ($value, $key) {
        return isset($value['command']) && isset($value['model_uuid']) && isset($value['occurred_at']) && $this->isDateStringValid($value['occurred_at']) && Uuid::isValid($value['model_uuid']) ? $value : null;
      });

    $validArray = $valid->toArray();
    StableSort::usort($validArray, function ($a, $b) {
      return new \DateTime($a['occurred_at']) <=> new \DateTime($b['occurred_at']);
    });
    $valid = collect($validArray);

    $valid = $valid->groupBy([function ($key, $value) {
      list($model, $command) = explode(':', $key['command']);
      return $model;
    }, 'model_uuid']);

    $errors = $invalid->map(function ($value, $key) {
      return $this->formatError($value, __('event-driven-service::events.invalid'), true);
    });

    $updateResponse = [];

    foreach ($valid as $modelType => $uuids) {
      $model = EventHelper::getModelByType($modelType);
      foreach ($uuids as $uuid => $events) {
        $currentModel = $model->withAll()->where($model->modelKeyField, $uuid)->first();
        $created = false;
        if (!$currentModel) {
          $created = true;
          $modelData = [
            $model->modelKeyField => $uuid,
          ];
          if ($model->modelKeyField != 'uuid') {
            $modelData['uuid'] = $model->generateUuid();
          }
          $currentModel = $model->make($modelData)->refresh();
        }
        if ($currentModel->trashed()) {
          $deletedEvents = collect($events)->map(function ($event) {
            return [
              'request' => $event,
              'error' => __('event-driven-service::events.deleted')
            ];
          });
          $errors = $errors->merge($deletedEvents);
          continue;
        }
        $lastEventIndex = $events->count() - 1;
        foreach ($events as $i => $event) {
          $event['is_bulk'] = $isBulk;
          $event['last_event'] = $i === $lastEventIndex;
          list($event['model'], $event['operation']) = explode(':', $event['command']);
          try {
            Validator::make($event, [
              'operation' => 'required|in:' . implode(",", $currentModel->getCommands())
            ])->validate();

            if ($event['operation'] == 'create') {
              if (!$created && !$currentModel->canRecreate()) {
                throw new BadRequestHttpException(__('event-driven-service::events.created'));
              }
              if (is_string($event['value'])) {
                $event['value'] = json_decode($event['value'], true);
              }
            }

            if (method_exists($this, $event['operation'])) {
              $response = call_user_func_array([$this, $event['operation']], [$currentModel, $event]);
            } else {
              $response = call_user_func_array([$currentModel, $event['operation']], [$event]);
            }

            if ($processUpdateResponse) {
              $updateResponse = array_merge($updateResponse, $currentModel->getUpdatesResponse());
            }

            if ($response) {
              $errors[] = $response;
            }
          } catch (ValidationException $e) {
            $this->logError($e);
            $errors[] = $this->handleValidationError($event, $e);
            continue;
          } catch (\Exception $e) {
            $this->logError($e);
            $errors[] = $this->formatError($event, $e->getMessage());
            continue;
          }
        }
      }
    }
    $this->failedEventRepository->createAll($errors);
    return [$errors, $updateResponse];
  }

  public function create(EventModel $model, $event) {
    try {
      Validator::make($event, [
        'value' => 'required'
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    $event['value']['uuid'] = $model->uuid;

    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->createModel($event);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  public function update(EventModel $model, $event) {

    try {
      Validator::make($event, [
        'name' => 'required|string',
        'value' => 'nullable',
        'version' => ['nullable', new UuidRule],
        'version_new' => ['required', new UuidRule]
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    $nonColumnField = in_array($event['name'], $model->getNonColumnFields());

    if (!in_array($event['name'], $model->getFillable()) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.not_found'));
    }

    if (!in_array($event['name'] . '_version', $model->getFillable()) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.offline'));
    }

    if (!$model->isCurrent($event['name'], $event['version']) && !$nonColumnField) {
      throw new BadRequestHttpException(__('event-driven-service::events.update.version'));
    }

    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->updateModel($event);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  public function delete(EventModel $model, $event) {
    try {
      DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->deleteModel($event);
      DB::commit();
    } catch (\Exception $e) {
      DB::rollBack();
      throw $e;
    }
  }

  private function isDateStringValid($date) {
    try {
      Carbon::createFromFormat(EventHelper::OCCURRED_AT_FORMAT, $date);
      return true;
    } catch (\Exception $e) {
      return false;
    }
  }
}